package output

import (
	"bytes"
	"errors"
	"fmt"
	"sync"
	"testing"
	"time"

	"github.com/Jeffail/benthos/v3/lib/condition"
	"github.com/Jeffail/benthos/v3/lib/message"
	"github.com/Jeffail/benthos/v3/lib/response"
	"github.com/Jeffail/benthos/v3/lib/types"
	"github.com/stretchr/testify/require"
)

func TestSwitchDeprecatedNoConditions(t *testing.T) {
	nOutputs, nMsgs := 10, 1000

	conf := NewConfig()
	mockOutputs := []*MockOutputType{}
	for i := 0; i < nOutputs; i++ {
		conf.Switch.Outputs = append(conf.Switch.Outputs, NewSwitchConfigOutput())
		conf.Switch.Outputs[i].Fallthrough = true
		mockOutputs = append(mockOutputs, &MockOutputType{})
	}

	s := newSwitch(t, conf, mockOutputs)

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	require.NoError(t, s.Consume(readChan))

	for i := 0; i < nMsgs; i++ {
		content := [][]byte{[]byte(fmt.Sprintf("hello world %v", i))}
		select {
		case readChan <- types.NewTransaction(message.New(content), resChan):
		case <-time.After(time.Second):
			t.Errorf("Timed out waiting for broker send")
			return
		}
		resChanSlice := []chan<- types.Response{}
		for j := 0; j < nOutputs; j++ {
			var ts types.Transaction
			select {
			case ts = <-mockOutputs[j].TChan:
				if !bytes.Equal(ts.Payload.Get(0).Get(), content[0]) {
					t.Errorf("Wrong content returned %s != %s", ts.Payload.Get(0).Get(), content[0])
				}
				resChanSlice = append(resChanSlice, ts.ResponseChan)
			case <-time.After(time.Second):
				t.Errorf("Timed out waiting for broker propagate")
				return
			}
		}
		for j := 0; j < nOutputs; j++ {
			select {
			case resChanSlice[j] <- response.NewAck():
			case <-time.After(time.Second):
				t.Errorf("Timed out responding to broker")
				return
			}
		}
		select {
		case res := <-resChan:
			if res.Error() != nil {
				t.Errorf("Received unexpected errors from broker: %v", res.Error())
			}
		case <-time.After(time.Second):
			t.Errorf("Timed out responding to broker")
			return
		}
	}

	s.CloseAsync()

	if err := s.WaitForClose(time.Second * 5); err != nil {
		t.Error(err)
	}
}

func TestSwitchDeprecatedNoRetries(t *testing.T) {
	nOutputs, nMsgs := 10, 1000

	conf := NewConfig()
	conf.Switch.RetryUntilSuccess = false
	mockOutputs := []*MockOutputType{}
	for i := 0; i < nOutputs; i++ {
		conf.Switch.Outputs = append(conf.Switch.Outputs, NewSwitchConfigOutput())
		conf.Switch.Outputs[i].Fallthrough = true
		mockOutputs = append(mockOutputs, &MockOutputType{})
	}

	s := newSwitch(t, conf, mockOutputs)

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	require.NoError(t, s.Consume(readChan))

	for i := 0; i < nMsgs; i++ {
		content := [][]byte{[]byte(fmt.Sprintf("hello world %v", i))}
		select {
		case readChan <- types.NewTransaction(message.New(content), resChan):
		case <-time.After(time.Second):
			t.Fatal("Timed out waiting for broker send")
		}
		resChanSlice := []chan<- types.Response{}
		for j := 0; j < nOutputs; j++ {
			var ts types.Transaction
			select {
			case ts = <-mockOutputs[j].TChan:
				if !bytes.Equal(ts.Payload.Get(0).Get(), content[0]) {
					t.Errorf("Wrong content returned %s != %s", ts.Payload.Get(0).Get(), content[0])
				}
				resChanSlice = append(resChanSlice, ts.ResponseChan)
			case <-time.After(time.Second):
				t.Fatal("Timed out waiting for broker propagate")
			}
		}
		for j := 0; j < nOutputs; j++ {
			var res types.Response
			if j == 1 {
				res = response.NewError(errors.New("test"))
			} else {
				res = response.NewAck()
			}
			select {
			case resChanSlice[j] <- res:
			case <-time.After(time.Second):
				t.Fatal("Timed out responding to broker")
			}
		}
		select {
		case res := <-resChan:
			if res.Error() == nil {
				t.Error("Received nil error from broker")
			} else if exp, act := "test", res.Error().Error(); exp != act {
				t.Errorf("Wrong error message from broker: %v != %v", act, exp)
			}
		case <-time.After(time.Second):
			t.Fatal("Timed out responding to broker")
		}
	}

	s.CloseAsync()

	if err := s.WaitForClose(time.Second * 5); err != nil {
		t.Error(err)
	}
}

func TestSwitchDeprecatedWithConditions(t *testing.T) {
	nMsgs := 100

	mockOutputs := []*MockOutputType{{}, {}, {}}

	conf := NewConfig()
	for i := 0; i < len(mockOutputs); i++ {
		conf.Switch.Outputs = append(conf.Switch.Outputs, NewSwitchConfigOutput())
	}

	fooConfig := condition.NewConfig()
	fooConfig.Type = condition.TypeJMESPath
	fooConfig.JMESPath.Query = "foo == 'bar'"
	conf.Switch.Outputs[0].Condition = fooConfig

	barConfig := condition.NewConfig()
	barConfig.Type = condition.TypeJMESPath
	barConfig.JMESPath.Query = "foo == 'baz'"
	conf.Switch.Outputs[1].Condition = barConfig

	s := newSwitch(t, conf, mockOutputs)

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	require.NoError(t, s.Consume(readChan))

	wg := sync.WaitGroup{}
	wg.Add(1)
	go func() {
		defer wg.Done()
		closed := 0
		bar := `{"foo":"bar"}`
		baz := `{"foo":"baz"}`

	outputLoop:
		for closed < len(mockOutputs) {
			var ts types.Transaction
			var ok bool

			select {
			case ts, ok = <-mockOutputs[0].TChan:
				if !ok {
					closed++
					continue outputLoop
				}
				if act := string(ts.Payload.Get(0).Get()); act != bar {
					t.Errorf("Expected output 0 msgs to equal %s, got %s", bar, act)
				}
			case ts, ok = <-mockOutputs[1].TChan:
				if !ok {
					closed++
					continue outputLoop
				}
				if act := string(ts.Payload.Get(0).Get()); act != baz {
					t.Errorf("Expected output 1 msgs to equal %s, got %s", baz, act)
				}
			case ts, ok = <-mockOutputs[2].TChan:
				if !ok {
					closed++
					continue outputLoop
				}
				if act := string(ts.Payload.Get(0).Get()); act == bar || act == baz {
					t.Errorf("Expected output 2 msgs to not equal %s or %s, got %s", bar, baz, act)
				}
			case <-time.After(time.Second):
				t.Error("Timed out waiting for output to propagate")
				break outputLoop
			}

			select {
			case ts.ResponseChan <- response.NewAck():
			case <-time.After(time.Second):
				t.Errorf("Timed out responding to output")
				break outputLoop
			}
		}
	}()

	for i := 0; i < nMsgs; i++ {
		foo := "bar"
		if i%3 == 0 {
			foo = "qux"
		} else if i%2 == 0 {
			foo = "baz"
		}
		content := [][]byte{[]byte(fmt.Sprintf("{\"foo\":%q}", foo))}
		select {
		case readChan <- types.NewTransaction(message.New(content), resChan):
		case <-time.After(time.Second):
			t.Errorf("Timed out waiting for output send")
			return
		}

		select {
		case res := <-resChan:
			if res.Error() != nil {
				t.Errorf("Received unexpected errors from output: %v", res.Error())
			}
		case <-time.After(time.Second):
			t.Errorf("Timed out responding to output")
			return
		}
	}

	s.CloseAsync()
	if err := s.WaitForClose(time.Second * 5); err != nil {
		t.Error(err)
	}
	wg.Wait()
}

func TestSwitchDeprecatedNoMatch(t *testing.T) {
	mockOutputs := []*MockOutputType{{}, {}, {}}

	conf := NewConfig()
	for i := 0; i < len(mockOutputs); i++ {
		conf.Switch.Outputs = append(conf.Switch.Outputs, NewSwitchConfigOutput())
	}

	fooConfig := condition.NewConfig()
	fooConfig.Type = condition.TypeJMESPath
	fooConfig.JMESPath.Query = "foo == 'bar'"
	conf.Switch.Outputs[0].Condition = fooConfig

	barConfig := condition.NewConfig()
	barConfig.Type = condition.TypeJMESPath
	barConfig.JMESPath.Query = "foo == 'baz'"
	conf.Switch.Outputs[1].Condition = barConfig

	bazConfig := condition.NewConfig()
	bazConfig.Type = condition.TypeStatic
	bazConfig.Static = false
	conf.Switch.Outputs[2].Condition = bazConfig

	s := newSwitch(t, conf, mockOutputs)

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	require.NoError(t, s.Consume(readChan))

	msg := message.New([][]byte{[]byte(`{"foo":"qux"}`)})
	select {
	case readChan <- types.NewTransaction(msg, resChan):
	case <-time.After(time.Second):
		t.Errorf("Timed out waiting for output send")
		return
	}

	select {
	case res := <-resChan:
		if err := res.Error(); err != nil {
			t.Errorf("Expected error from output to equal nil, got %v", err)
		}
	case <-time.After(time.Second):
		t.Errorf("Timed out responding to output")
		return
	}

	s.CloseAsync()

	if err := s.WaitForClose(time.Second * 5); err != nil {
		t.Error(err)
	}
}

func TestSwitchDeprecatedNoMatchStrict(t *testing.T) {
	mockOutputs := []*MockOutputType{{}, {}, {}}

	conf := NewConfig()
	conf.Switch.StrictMode = true
	for i := 0; i < len(mockOutputs); i++ {
		conf.Switch.Outputs = append(conf.Switch.Outputs, NewSwitchConfigOutput())
	}

	fooConfig := condition.NewConfig()
	fooConfig.Type = condition.TypeJMESPath
	fooConfig.JMESPath.Query = "foo == 'bar'"
	conf.Switch.Outputs[0].Condition = fooConfig

	barConfig := condition.NewConfig()
	barConfig.Type = condition.TypeJMESPath
	barConfig.JMESPath.Query = "foo == 'baz'"
	conf.Switch.Outputs[1].Condition = barConfig

	bazConfig := condition.NewConfig()
	bazConfig.Type = condition.TypeStatic
	bazConfig.Static = false
	conf.Switch.Outputs[2].Condition = bazConfig

	s := newSwitch(t, conf, mockOutputs)

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	require.NoError(t, s.Consume(readChan))

	msg := message.New([][]byte{[]byte(`{"foo":"qux"}`)})
	select {
	case readChan <- types.NewTransaction(msg, resChan):
	case <-time.After(time.Second):
		t.Errorf("Timed out waiting for output send")
		return
	}

	select {
	case res := <-resChan:
		if err := res.Error(); err == nil {
			t.Error("Expected error from output but was nil")
		}
	case <-time.After(time.Second):
		t.Errorf("Timed out responding to output")
		return
	}

	s.CloseAsync()

	if err := s.WaitForClose(time.Second * 5); err != nil {
		t.Error(err)
	}
}

func TestSwitchDeprecatedWithConditionsNoFallthrough(t *testing.T) {
	nMsgs := 100

	mockOutputs := []*MockOutputType{{}, {}, {}}

	conf := NewConfig()
	for i := 0; i < len(mockOutputs); i++ {
		conf.Switch.Outputs = append(conf.Switch.Outputs, NewSwitchConfigOutput())
	}

	fooConfig := condition.NewConfig()
	fooConfig.Type = condition.TypeJMESPath
	fooConfig.JMESPath.Query = "foo == 'bar'"
	conf.Switch.Outputs[0].Condition = fooConfig

	barConfig := condition.NewConfig()
	barConfig.Type = condition.TypeJMESPath
	barConfig.JMESPath.Query = "foo == 'baz'"
	conf.Switch.Outputs[1].Condition = barConfig

	s := newSwitch(t, conf, mockOutputs)

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	require.NoError(t, s.Consume(readChan))

	wg := sync.WaitGroup{}
	wg.Add(1)
	go func() {
		defer wg.Done()

		closed := 0
		bar := `{"foo":"bar"}`
		baz := `{"foo":"baz"}`

	outputLoop:
		for closed < len(mockOutputs) {
			var ts types.Transaction
			var ok bool

			resChans := []chan<- types.Response{}
			for len(resChans) < 1 {
				select {
				case ts, ok = <-mockOutputs[0].TChan:
					if !ok {
						closed++
						continue outputLoop
					}
					if act := string(ts.Payload.Get(0).Get()); act != bar {
						t.Errorf("Expected output 0 msgs to equal %s, got %s", bar, act)
					}
				case ts, ok = <-mockOutputs[1].TChan:
					if !ok {
						closed++
						continue outputLoop
					}
					if act := string(ts.Payload.Get(0).Get()); act != baz {
						t.Errorf("Expected output 1 msgs to equal %s, got %s", baz, act)
					}
				case ts, ok = <-mockOutputs[2].TChan:
					if !ok {
						closed++
						continue outputLoop
					}
					t.Error("Unexpected msg received by output 3")
				case <-time.After(time.Second):
					t.Error("Timed out waiting for output to propagate")
					break outputLoop
				}
				if ts.ResponseChan != nil {
					resChans = append(resChans, ts.ResponseChan)
				}
			}

			for i := 0; i < len(resChans); i++ {
				select {
				case resChans[i] <- response.NewAck():
				case <-time.After(time.Second):
					t.Errorf("Timed out responding to output")
					break outputLoop
				}
			}
		}
	}()

	for i := 0; i < nMsgs; i++ {
		foo := "bar"
		if i%2 == 0 {
			foo = "baz"
		}
		content := [][]byte{[]byte(fmt.Sprintf("{\"foo\":%q}", foo))}
		select {
		case readChan <- types.NewTransaction(message.New(content), resChan):
		case <-time.After(time.Second):
			t.Errorf("Timed out waiting for output send")
			return
		}

		select {
		case res := <-resChan:
			if res.Error() != nil {
				t.Errorf("Received unexpected errors from output: %v", res.Error())
			}
		case <-time.After(time.Second):
			t.Errorf("Timed out responding to output")
			return
		}
	}

	s.CloseAsync()

	if err := s.WaitForClose(time.Second * 5); err != nil {
		t.Error(err)
	}

	wg.Wait()
}

func TestSwitchDeprecatedAtLeastOnce(t *testing.T) {
	mockOne := MockOutputType{}
	mockTwo := MockOutputType{}

	mockOutputs := []*MockOutputType{
		&mockOne, &mockTwo,
	}

	conf := NewConfig()
	for i := 0; i < len(mockOutputs); i++ {
		outConf := NewSwitchConfigOutput()
		outConf.Fallthrough = true
		conf.Switch.Outputs = append(conf.Switch.Outputs, outConf)
	}

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	s := newSwitch(t, conf, mockOutputs)

	require.NoError(t, s.Consume(readChan))
	require.Error(t, s.Consume(readChan))

	select {
	case readChan <- types.NewTransaction(message.New([][]byte{[]byte("hello world")}), resChan):
	case <-time.After(time.Second):
		t.Error("Timed out waiting for output send")
		return
	}
	var ts1, ts2 types.Transaction
	select {
	case ts1 = <-mockOne.TChan:
	case <-time.After(time.Second):
		t.Error("Timed out waiting for mockOne")
		return
	}
	select {
	case ts2 = <-mockTwo.TChan:
	case <-time.After(time.Second):
		t.Error("Timed out waiting for mockOne")
		return
	}
	select {
	case ts1.ResponseChan <- response.NewAck():
	case <-time.After(time.Second):
		t.Error("Timed out responding to output")
		return
	}
	select {
	case ts2.ResponseChan <- response.NewError(errors.New("this is a test")):
	case <-time.After(time.Second):
		t.Error("Timed out responding to output")
		return
	}
	select {
	case <-mockOne.TChan:
		t.Error("Received duplicate message to mockOne")
	case ts2 = <-mockTwo.TChan:
	case <-resChan:
		t.Error("Received premature response from output")
	case <-time.After(time.Second):
		t.Error("Timed out waiting for mockTwo")
		return
	}
	select {
	case ts2.ResponseChan <- response.NewAck():
	case <-time.After(time.Second):
		t.Error("Timed out responding to output")
		return
	}
	select {
	case res := <-resChan:
		if res.Error() != nil {
			t.Errorf("Fan out returned error %v", res.Error())
		}
	case <-time.After(time.Second):
		t.Errorf("Timed out responding to output")
		return
	}

	close(readChan)

	if err := s.WaitForClose(time.Second * 5); err != nil {
		t.Error(err)
	}
}

func TestSwitchDeprecatedShutDownFromErrorResponse(t *testing.T) {
	mockOutputs := []*MockOutputType{{}, {}}

	conf := NewConfig()
	for i := 0; i < len(mockOutputs); i++ {
		outConf := NewSwitchConfigOutput()
		outConf.Fallthrough = true
		conf.Switch.Outputs = append(conf.Switch.Outputs, outConf)
	}

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	s := newSwitch(t, conf, mockOutputs)
	require.NoError(t, s.Consume(readChan))

	select {
	case readChan <- types.NewTransaction(message.New(nil), resChan):
	case <-time.After(time.Second):
		t.Error("Timed out waiting for msg send")
	}

	var ts types.Transaction
	var open bool
	select {
	case ts, open = <-mockOutputs[0].TChan:
		if !open {
			t.Error("Switch output closed early")
		}
	case <-time.After(time.Second):
		t.Error("Timed out waiting for msg rcv")
	}
	select {
	case _, open = <-mockOutputs[1].TChan:
		if !open {
			t.Error("Switch output closed early")
		}
	case <-time.After(time.Second):
		t.Error("Timed out waiting for msg rcv")
	}

	select {
	case ts.ResponseChan <- response.NewError(errors.New("test")):
	case <-time.After(time.Second):
		t.Error("Timed out waiting for res send")
	}

	s.CloseAsync()
	if err := s.WaitForClose(time.Second); err != nil {
		t.Error(err)
	}

	select {
	case _, open := <-mockOutputs[0].TChan:
		if open {
			t.Error("Switch output still open after closure")
		}
	case <-time.After(time.Second):
		t.Error("Timed out waiting for msg rcv")
	}
}

func TestSwitchDeprecatedShutDownFromReceive(t *testing.T) {
	mockOutputs := []*MockOutputType{{}, {}}

	conf := NewConfig()
	for i := 0; i < len(mockOutputs); i++ {
		outConf := NewSwitchConfigOutput()
		outConf.Fallthrough = true
		conf.Switch.Outputs = append(conf.Switch.Outputs, outConf)
	}

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	s := newSwitch(t, conf, mockOutputs)
	require.NoError(t, s.Consume(readChan))

	select {
	case readChan <- types.NewTransaction(message.New(nil), resChan):
	case <-time.After(time.Second):
		t.Error("Timed out waiting for msg send")
	}

	select {
	case _, open := <-mockOutputs[0].TChan:
		if !open {
			t.Error("Switch output closed early")
		}
	case <-time.After(time.Second):
		t.Error("Timed out waiting for msg rcv")
	}

	s.CloseAsync()
	if err := s.WaitForClose(time.Second); err != nil {
		t.Error(err)
	}

	select {
	case _, open := <-mockOutputs[0].TChan:
		if open {
			t.Error("Switch output still open after closure")
		}
	case <-time.After(time.Second):
		t.Error("Timed out waiting for msg rcv")
	}
}

func TestSwitchDeprecatedShutDownFromSend(t *testing.T) {
	mockOutputs := []*MockOutputType{{}, {}}

	conf := NewConfig()
	for i := 0; i < len(mockOutputs); i++ {
		outConf := NewSwitchConfigOutput()
		outConf.Fallthrough = true
		conf.Switch.Outputs = append(conf.Switch.Outputs, outConf)
	}

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	s := newSwitch(t, conf, mockOutputs)
	require.NoError(t, s.Consume(readChan))

	select {
	case readChan <- types.NewTransaction(message.New(nil), resChan):
	case <-time.After(time.Second):
		t.Error("Timed out waiting for msg send")
	}

	s.CloseAsync()
	if err := s.WaitForClose(time.Second); err != nil {
		t.Error(err)
	}

	select {
	case _, open := <-mockOutputs[0].TChan:
		if open {
			t.Error("Switch output still open after closure")
		}
	case <-time.After(time.Second):
		t.Error("Timed out waiting for msg rcv")
	}
}

func TestSwitchDeprecatedBackPressure(t *testing.T) {
	t.Parallel()

	mockOutputs := []*MockOutputType{{}, {}}

	conf := NewConfig()
	for i := 0; i < len(mockOutputs); i++ {
		outConf := NewSwitchConfigOutput()
		outConf.Fallthrough = true
		conf.Switch.Outputs = append(conf.Switch.Outputs, outConf)
	}

	readChan := make(chan types.Transaction)
	resChan := make(chan types.Response)

	s := newSwitch(t, conf, mockOutputs)
	require.NoError(t, s.Consume(readChan))

	wg := sync.WaitGroup{}
	wg.Add(1)
	doneChan := make(chan struct{})
	go func() {
		defer wg.Done()
		// Consume as fast as possible from mock one
		for {
			select {
			case ts := <-mockOutputs[0].TChan:
				select {
				case ts.ResponseChan <- response.NewAck():
				case <-doneChan:
					return
				}
			case <-doneChan:
				return
			}
		}
	}()

	i := 0
bpLoop:
	for ; i < 1000; i++ {
		select {
		case readChan <- types.NewTransaction(message.New([][]byte{[]byte("hello world")}), resChan):
		case <-time.After(time.Millisecond * 200):
			break bpLoop
		}
	}
	if i > 500 {
		t.Error("We shouldn't be capable of dumping this many messages into a blocked broker")
	}

	close(readChan)
	close(doneChan)
	wg.Wait()
}

//------------------------------------------------------------------------------
